@@ -1,7 +1,7 @@
 /*
  * Copyright 2000, International Business Machines Corporation and others.
  * All Rights Reserved.
- * 
+ *
  * This software has been released under the terms of the IBM Public
  * License.  For details, see the LICENSE file in the top-level source
  * directory or online at http://www.openafs.org/dl/license10.html
@@ -52,11 +52,7 @@
 #undef SHARED
 #include <rx/xdr.h>
 #include <afs/nfs.h>
-#ifdef AFS_PTHREAD_ENV
-#include <assert.h>
-#else /* AFS_PTHREAD_ENV */
-#include <afs/assert.h>
-#endif /* AFS_PTHREAD_ENV */
+#include <afs/afs_assert.h>
 #include <lwp.h>
 #include <lock.h>
 #include <afs/ptclient.h>
@@ -122,7 +118,6 @@ static afs_int32 Do_VLRegisterRPC(void);
 
 int eventlog = 0, rxlog = 0;
 FILE *debugFile;
-FILE *console = NULL;
 
 #ifdef AFS_PTHREAD_ENV
 pthread_mutex_t fsync_glock_mutex;
@@ -132,8 +127,7 @@ char fsync_wait[1];
 #endif /* AFS_PTHREAD_ENV */
 
 #ifdef AFS_NT40_ENV
-#define AFS_QUIETFS_ENV 1
-#define NT_OPEN_MAX    1024	/* This is an arbitrary no. we came up with for 
+#define NT_OPEN_MAX    1024	/* This is an arbitrary no. we came up with for
 				 * now. We hope this will be replaced by a more
 				 * intelligent estimate later. */
 #endif
@@ -164,9 +158,9 @@ int debuglevel = 0;
 int printBanner = 0;
 int rxJumbograms = 0;		/* default is to not send and receive jumbograms. */
 int rxBind = 0;		/* don't bind */
-int rxkadDisableDotCheck = 0;      /* disable check for dot in principal name */ 
+int rxkadDisableDotCheck = 0;      /* disable check for dot in principal name */
 int rxMaxMTU = -1;
-afs_int32 implicitAdminRights = PRSFS_LOOKUP;	/* The ADMINISTER right is 
+afs_int32 implicitAdminRights = PRSFS_LOOKUP;	/* The ADMINISTER right is
 						 * already implied */
 afs_int32 readonlyServer = 0;
 
@@ -203,6 +197,7 @@ int abort_threshold = 10;
 int udpBufSize = 0;		/* UDP buffer size for receive */
 int sendBufSize = 16384;	/* send buffer size */
 int saneacls = 0;		/* Sane ACLs Flag */
+static int unsafe_attach = 0;   /* avoid inUse check on vol attach? */
 
 struct timeval tp;
 
@@ -210,10 +205,6 @@ struct timeval tp;
 pthread_key_t viced_uclient_key;
 #endif
 
-#ifdef AFS_PTHREAD_ENV
-pthread_key_t viced_uclient_key;
-#endif
-
 /*
  * FileServer's name and IP address, both network byte order and
  * host byte order.
@@ -243,16 +234,37 @@ static void FlagMsg(void);
 #define PTHREAD_RWLOCK_INITIALIZER {0x2DA8B3B4, {0}}
 #endif
 
-struct fs_state fs_state = 
-    { FS_MODE_NORMAL, 
-      0, 
-      0, 
-      0, 
+#ifndef AFS_NT40_ENV
+struct fs_state fs_state =
+    { FS_MODE_NORMAL,
+      0,
+      0,
+      0,
       0,
       { 1,1,1,1 },
       PTHREAD_COND_INITIALIZER,
       PTHREAD_RWLOCK_INITIALIZER
     };
+#else /* AFS_NT40_ENV */
+struct fs_state fs_state;
+
+static int fs_stateInit(void)
+{
+    fs_state.mode = FS_MODE_NORMAL;
+    fs_state.FiveMinuteLWP_tranquil = 0;
+    fs_state.HostCheckLWP_tranquil = 0;
+    fs_state.FsyncCheckLWP_tranquil = 0;
+    fs_state.salvsync_fatal_error = 0;
+
+    fs_state.options.fs_state_save = 1;
+    fs_state.options.fs_state_restore = 1;
+    fs_state.options.fs_state_verify_before_save = 1;
+    fs_state.options.fs_state_verify_after_restore = 1;
+
+    CV_INIT(&fs_state.worker_done_cv, "worker done", CV_DEFAULT, 0);
+    osi_Assert(pthread_rwlock_init(&fs_state.state_lock, NULL) == 0);
+}
+#endif /* AFS_NT40_ENV */
 #endif /* AFS_DEMAND_ATTACH_FS */
 
 /*
@@ -265,8 +277,8 @@ CheckDescriptors(void *unused)
 {
 #ifndef AFS_NT40_ENV
     struct afs_stat status;
-    register int tsize = getdtablesize();
-    register int i;
+    int tsize = getdtablesize();
+    int i;
     for (i = 0; i < tsize; i++) {
 	if (afs_fstat(i, &status) != -1) {
 	    printf("%d: dev %x, inode %u, length %u, type/mode %x\n", i,
@@ -321,9 +333,9 @@ CheckDescriptors_Signal(int x)
 }
 #endif /* AFS_PTHREAD_ENV */
 
-/* check whether caller is authorized to manage RX statistics */
+/* check whether caller is authorized to perform admin operations */
 int
-fs_rxstat_userok(struct rx_call *call)
+viced_SuperUser(struct rx_call *call)
 {
     return afsconf_SuperUser(confDir, call, NULL);
 }
@@ -480,15 +492,9 @@ FiveMinuteCheckLWP(void *unused)
 #endif
 	if (printBanner && (++msg & 1)) {	/* Every 10 minutes */
 	    time_t now = FT_ApproxTime();
-	    if (console != NULL) {
-#ifndef AFS_QUIETFS_ENV
-		fprintf(console, "File server is running at %s\r",
-			afs_ctime(&now, tbuffer, sizeof(tbuffer)));
-#endif /* AFS_QUIETFS_ENV */
-		ViceLog(2,
-			("File server is running at %s\n",
-			 afs_ctime(&now, tbuffer, sizeof(tbuffer))));
-	    }
+	    ViceLog(2,
+		    ("File server is running at %s\n",
+		     afs_ctime(&now, tbuffer, sizeof(tbuffer))));
 	}
 #ifdef AFS_DEMAND_ATTACH_FS
 	FS_STATE_WRLOCK;
@@ -497,7 +503,7 @@ FiveMinuteCheckLWP(void *unused)
 #ifdef AFS_DEMAND_ATTACH_FS
     fs_state.FiveMinuteLWP_tranquil = 1;
     FS_LOCK;
-    assert(pthread_cond_broadcast(&fs_state.worker_done_cv)==0);
+    CV_BROADCAST(&fs_state.worker_done_cv);
     FS_UNLOCK;
     FS_STATE_UNLOCK;
 #endif
@@ -549,7 +555,7 @@ HostCheckLWP(void *unused)
 #ifdef AFS_DEMAND_ATTACH_FS
     fs_state.HostCheckLWP_tranquil = 1;
     FS_LOCK;
-    assert(pthread_cond_broadcast(&fs_state.worker_done_cv)==0);
+    CV_BROADCAST(&fs_state.worker_done_cv);
     FS_UNLOCK;
     FS_STATE_UNLOCK;
 #endif
@@ -585,9 +591,8 @@ FsyncCheckLWP(void *unused)
 	fsync_next.tv_nsec = 0;
 	fsync_next.tv_sec = time(0) + fiveminutes;
 
-	code =
-	    pthread_cond_timedwait(&fsync_cond, &fsync_glock_mutex,
-				   &fsync_next);
+	code = CV_TIMEDWAIT(&fsync_cond, &fsync_glock_mutex,
+			    &fsync_next);
 	if (code != 0 && code != ETIMEDOUT)
 	    ViceLog(0, ("pthread_cond_timedwait returned %d\n", code));
 #else /* AFS_PTHREAD_ENV */
@@ -616,7 +621,7 @@ FsyncCheckLWP(void *unused)
 #ifdef AFS_DEMAND_ATTACH_FS
     fs_state.FsyncCheckLWP_tranquil = 1;
     FS_LOCK;
-    assert(pthread_cond_broadcast(&fs_state.worker_done_cv)==0);
+    CV_BROADCAST(&fs_state.worker_done_cv);
     FS_UNLOCK;
     FS_STATE_UNLOCK;
 #endif /* AFS_DEMAND_ATTACH_FS */
@@ -702,6 +707,9 @@ PrintCounters(void)
     int workstations, activeworkstations, delworkstations;
     int processSize = 0;
     char tbuffer[32];
+#ifdef AFS_DEMAND_ATTACH_FS
+    int stats_flags = 0;
+#endif
 
     FT_GetTimeOfDay(&tpl, 0);
     Statistics = 1;
@@ -710,9 +718,12 @@ PrintCounters(void)
 	     afs_ctime(&StartTime, tbuffer, sizeof(tbuffer))));
 
 #ifdef AFS_DEMAND_ATTACH_FS
-    /* XXX perhaps set extended stats verbosity flags
-     * based upon LogLevel ?? */
-    VPrintExtendedCacheStats(VOL_STATS_PER_CHAIN2);
+    if (LogLevel >= 125) {
+	stats_flags = VOL_STATS_PER_CHAIN2;
+    } else if (LogLevel >= 25) {
+	stats_flags = VOL_STATS_PER_CHAIN;
+    }
+    VPrintExtendedCacheStats(stats_flags);
 #endif
     VPrintCacheStats();
     VPrintDiskStats();
@@ -770,7 +781,8 @@ ShutdownWatchdogLWP(void *unused)
     sleep(panic_timeout);
     ViceLog(0, ("ShutdownWatchdogLWP: Failed to shutdown and panic "
                 "within %d seconds; forcing panic\n", panic_timeout));
-    assert(0);
+    osi_Panic("ShutdownWatchdogLWP: Failed to shutdown and panic "
+	      "within %d seconds; forcing panic\n", panic_timeout);
     return NULL;
 }
 
@@ -784,11 +796,11 @@ ShutDownAndCore(int dopanic)
 #ifdef AFS_PTHREAD_ENV
 	pthread_t watchdogPid;
 	pthread_attr_t tattr;
-	assert(pthread_attr_init(&tattr) == 0);
-	assert(pthread_create(&watchdogPid, &tattr, ShutdownWatchdogLWP, NULL) == 0);
+	osi_Assert(pthread_attr_init(&tattr) == 0);
+	osi_Assert(pthread_create(&watchdogPid, &tattr, ShutdownWatchdogLWP, NULL) == 0);
 #else
 	PROCESS watchdogPid;
-	assert(LWP_CreateProcess
+	osi_Assert(LWP_CreateProcess
 	       (ShutdownWatchdogLWP, stack * 1024, LWP_MAX_PRIORITY - 2,
 	        NULL, "ShutdownWatchdog", &watchdogPid) == LWP_SUCCESS);
 #endif
@@ -820,12 +832,6 @@ ShutDownAndCore(int dopanic)
 	     afs_ctime(&now, tbuffer, sizeof(tbuffer))));
     if (dopanic)
 	ViceLog(0, ("ABNORMAL SHUTDOWN, see core file.\n"));
-#ifndef AFS_QUIETFS_ENV
-    if (console != NULL) {
-	fprintf(console, "File server restart/shutdown received at %s\r",
-		afs_ctime(&now, tbuffer, sizeof(tbuffer)));
-    }
-#endif
     DFlush();
     if (!dopanic)
 	PrintCounters();
@@ -835,7 +841,7 @@ ShutDownAndCore(int dopanic)
 
 #ifdef AFS_DEMAND_ATTACH_FS
     if (fs_state.options.fs_state_save) {
-	/* 
+	/*
 	 * demand attach fs
 	 * save fileserver state to disk */
 
@@ -852,7 +858,7 @@ ShutDownAndCore(int dopanic)
 		FS_LOCK;
 		FS_STATE_UNLOCK;
 		ViceLog(0, ("waiting for background host/callback threads to quiesce before saving fileserver state...\n"));
-		assert(pthread_cond_wait(&fs_state.worker_done_cv, &fileproc_glock_mutex) == 0);
+		CV_WAIT(&fs_state.worker_done_cv, &fileproc_glock_mutex);
 		FS_UNLOCK;
 		FS_STATE_RDLOCK;
 	    }
@@ -870,29 +876,19 @@ ShutDownAndCore(int dopanic)
 	rx_PrintStats(debugFile);
 	fflush(debugFile);
     }
-    if (console != NULL) {
-	now = time(0);
-	if (dopanic) {
-#ifndef AFS_QUIETFS_ENV
-	    fprintf(console, "File server has terminated abnormally at %s\r",
-		    afs_ctime(&now, tbuffer, sizeof(tbuffer)));
-#endif
-	    ViceLog(0,
-		    ("File server has terminated abnormally at %s\n",
-		     afs_ctime(&now, tbuffer, sizeof(tbuffer))));
-	} else {
-#ifndef AFS_QUIETFS_ENV
-	    fprintf(console, "File server has terminated normally at %s\r",
-		    afs_ctime(&now, tbuffer, sizeof(tbuffer)));
-#endif
-	    ViceLog(0,
-		    ("File server has terminated normally at %s\n",
-		     afs_ctime(&now, tbuffer, sizeof(tbuffer))));
-	}
+    now = time(0);
+    if (dopanic) {
+      ViceLog(0,
+	      ("File server has terminated abnormally at %s\n",
+	       afs_ctime(&now, tbuffer, sizeof(tbuffer))));
+    } else {
+      ViceLog(0,
+	      ("File server has terminated normally at %s\n",
+	       afs_ctime(&now, tbuffer, sizeof(tbuffer))));
     }
 
-    if (dopanic)
-	assert(0);
+    if (dopanic) /* XXX pass in file and line? */
+	osi_Panic("Panic requested\n");
 
     exit(0);
 }
@@ -949,6 +945,7 @@ FlagMsg(void)
     fputs("[-vlruthresh <minutes before unused volumes become eligible for soft detach> (default is 2 hours)] ", stdout);
     fputs("[-vlruinterval <seconds between VLRU scans> (default is 2 minutes)] ", stdout);
     fputs("[-vlrumax <max volumes to soft detach in one VLRU scan> (default is 8)] ", stdout);
+    fputs("[-unsafe-nosalvage (bypass volume inUse safety check on attach, bypassing salvage)] ", stdout);
 #elif AFS_PTHREAD_ENV
     fputs("[-vattachpar <number of volume attach threads> (default is 1)] ", stdout);
 #endif
@@ -967,6 +964,7 @@ FlagMsg(void)
     fputs("[-abortthreshold <abort threshold>] ", stdout);
     fputs("[-nojumbo (disable jumbogram network packets - deprecated)] ", stdout);
     fputs("[-jumbo (enable jumbogram network packets)] ", stdout);
+    fputs("[-sync <always | delayed | onclose | never>]", stdout);
 /*   fputs("[-enable_peer_stats] ", stdout); */
 /*   fputs("[-enable_process_stats] ", stdout); */
     fputs("[-help]\n", stdout);
@@ -1084,13 +1082,14 @@ ParseArgs(int argc, char *argv[])
     int Sawbusy = 0;
     int i;
     int bufSize = 0;		/* temp variable to read in udp socket buf size */
+    int lwps_max;
     char *auditFileName = NULL;
 
     for (i = 1; i < argc; i++) {
 	if (!strcmp(argv[i], "-d")) {
             if ((i + 1) >= argc) {
-		fprintf(stderr, "missing argument for -d\n"); 
-		return -1; 
+		fprintf(stderr, "missing argument for -d\n");
+		return -1;
 	    }
 	    debuglevel = atoi(argv[++i]);
 	    LogLevel = debuglevel;
@@ -1098,8 +1097,8 @@ ParseArgs(int argc, char *argv[])
 	    printBanner = 1;
 	} else if (!strcmp(argv[i], "-implicit")) {
             if ((i + 1) >= argc) {
-		fprintf(stderr, "missing argument for -implicit\n"); 
-		return -1; 
+		fprintf(stderr, "missing argument for -implicit\n");
+		return -1;
 	    }
 	    implicitAdminRights = ParseRights(argv[++i]);
 	    if (implicitAdminRights < 0)
@@ -1111,37 +1110,31 @@ ParseArgs(int argc, char *argv[])
 	} else if (!strcmp(argv[i], "-S")) {
 	    SawS = 1;
 	} else if (!strcmp(argv[i], "-p")) {
-	    int lwps_max =
-		max_fileserver_thread() - FILESERVER_HELPER_THREADS;
 	    Sawlwps = 1;
             if ((i + 1) >= argc) {
-		fprintf(stderr, "missing argument for -p\n"); 
-		return -1; 
+		fprintf(stderr, "missing argument for -p\n");
+		return -1;
 	    }
 	    lwps = atoi(argv[++i]);
-	    if (lwps > lwps_max)
-		lwps = lwps_max;
-	    else if (lwps < 6)
-		lwps = 6;
 	} else if (!strcmp(argv[i], "-b")) {
 	    Sawbufs = 1;
             if ((i + 1) >= argc) {
-		fprintf(stderr, "missing argument for -b\n"); 
-		return -1; 
+		fprintf(stderr, "missing argument for -b\n");
+		return -1;
 	    }
 	    buffs = atoi(argv[++i]);
 	} else if (!strcmp(argv[i], "-l")) {
 	    Sawlarge = 1;
             if ((i + 1) >= argc) {
-		fprintf(stderr, "missing argument for -l\n"); 
-		return -1; 
+		fprintf(stderr, "missing argument for -l\n");
+		return -1;
 	    }
 	    large = atoi(argv[++i]);
 	} else if (!strcmp(argv[i], "-vc")) {
 	    SawVC = 1;
             if ((i + 1) >= argc) {
-		fprintf(stderr, "missing argument for -vc\n"); 
-		return -1; 
+		fprintf(stderr, "missing argument for -vc\n");
+		return -1;
 	    }
 	    volcache = atoi(argv[++i]);
 	} else if (!strcmp(argv[i], "-novbc")) {
@@ -1149,15 +1142,15 @@ ParseArgs(int argc, char *argv[])
 	} else if (!strcmp(argv[i], "-rxpck")) {
 	    Sawrxpck = 1;
             if ((i + 1) >= argc) {
-		fprintf(stderr, "missing argument for -rxpck\n"); 
-		return -1; 
+		fprintf(stderr, "missing argument for -rxpck\n");
+		return -1;
 	    }
 	    rxpackets = atoi(argv[++i]);
 #ifdef AFS_PTHREAD_ENV
 	} else if (!strcmp(argv[i], "-vattachpar")) {
             if ((i + 1) >= argc) {
-		fprintf(stderr, "missing argument for %s\n", argv[i]); 
-		return -1; 
+		fprintf(stderr, "missing argument for %s\n", argv[i]);
+		return -1;
 	    }
 	    vol_attach_threads = atoi(argv[++i]);
 #endif /* AFS_PTHREAD_ENV */
@@ -1186,8 +1179,8 @@ ParseArgs(int argc, char *argv[])
 	    fs_state.options.fs_state_restore = 0;
 	} else if (!strcmp(argv[i], "-fs-state-verify")) {
             if ((i + 1) >= argc) {
-		fprintf(stderr, "missing argument for %s\n", argv[i]); 
-		return -1; 
+		fprintf(stderr, "missing argument for %s\n", argv[i]);
+		return -1;
 	    }
 	    i++;
 	    if (!strcmp(argv[i], "none")) {
@@ -1204,49 +1197,57 @@ ParseArgs(int argc, char *argv[])
 		return -1;
 	    }
 	} else if (!strcmp(argv[i], "-vhashsize")) {
+	    int hashsize;
             if ((i + 1) >= argc) {
-		fprintf(stderr, "missing argument for %s\n", argv[i]); 
-		return -1; 
+		fprintf(stderr, "missing argument for %s\n", argv[i]);
+		return -1;
+	    }
+	    hashsize = atoi(argv[++i]);
+	    if (VSetVolHashSize(hashsize)) {
+		fprintf(stderr, "specified -vhashsize (%s) is invalid or out "
+		                "of range\n", argv[i]);
+		return -1;
 	    }
-	    VSetVolHashSize(atoi(argv[++i]));
 	} else if (!strcmp(argv[i], "-vlrudisable")) {
 	    VLRU_SetOptions(VLRU_SET_ENABLED, 0);
 	} else if (!strcmp(argv[i], "-vlruthresh")) {
             if ((i + 1) >= argc) {
-		fprintf(stderr, "missing argument for %s\n", argv[i]); 
-		return -1; 
+		fprintf(stderr, "missing argument for %s\n", argv[i]);
+		return -1;
 	    }
 	    VLRU_SetOptions(VLRU_SET_THRESH, 60*atoi(argv[++i]));
 	} else if (!strcmp(argv[i], "-vlruinterval")) {
             if ((i + 1) >= argc) {
-		fprintf(stderr, "missing argument for %s\n", argv[i]); 
-		return -1; 
+		fprintf(stderr, "missing argument for %s\n", argv[i]);
+		return -1;
 	    }
 	    VLRU_SetOptions(VLRU_SET_INTERVAL, atoi(argv[++i]));
 	} else if (!strcmp(argv[i], "-vlrumax")) {
             if ((i + 1) >= argc) {
-		fprintf(stderr, "missing argument for %s\n", argv[i]); 
-		return -1; 
+		fprintf(stderr, "missing argument for %s\n", argv[i]);
+		return -1;
 	    }
 	    VLRU_SetOptions(VLRU_SET_MAX, atoi(argv[++i]));
+	} else if (!strcmp(argv[i], "-unsafe-nosalvage")) {
+	    unsafe_attach = 1;
 #endif /* AFS_DEMAND_ATTACH_FS */
 	} else if (!strcmp(argv[i], "-s")) {
 	    Sawsmall = 1;
             if ((i + 1) >= argc) {
-		fprintf(stderr, "missing argument for -s\n"); 
-		return -1; 
+		fprintf(stderr, "missing argument for -s\n");
+		return -1;
 	    }
 	    nSmallVns = atoi(argv[++i]);
 	} else if (!strcmp(argv[i], "-abortthreshold")) {
             if ((i + 1) >= argc) {
-		fprintf(stderr, "missing argument for -abortthreshold\n"); 
-		return -1; 
+		fprintf(stderr, "missing argument for -abortthreshold\n");
+		return -1;
 	    }
 	    abort_threshold = atoi(argv[++i]);
 	} else if (!strcmp(argv[i], "-k")) {
             if ((i + 1) >= argc) {
-		fprintf(stderr, "missing argument for -k\n"); 
-		return -1; 
+		fprintf(stderr, "missing argument for -k\n");
+		return -1;
 	    }
 	    stack = atoi(argv[++i]);
 	}
@@ -1257,30 +1258,30 @@ ParseArgs(int argc, char *argv[])
 #endif
 	else if (!strcmp(argv[i], "-spare")) {
             if ((i + 1) >= argc) {
-		fprintf(stderr, "missing argument for -spare\n"); 
-		return -1; 
+		fprintf(stderr, "missing argument for -spare\n");
+		return -1;
 	    }
 	    BlocksSpare = atoi(argv[++i]);
 	    SawSpare = 1;
 	} else if (!strcmp(argv[i], "-pctspare")) {
             if ((i + 1) >= argc) {
-		fprintf(stderr, "missing argument for -pctspare\n"); 
-		return -1; 
+		fprintf(stderr, "missing argument for -pctspare\n");
+		return -1;
 	    }
 	    PctSpare = atoi(argv[++i]);
 	    BlocksSpare = 0;	/* has non-zero default */
 	    SawPctSpare = 1;
 	} else if (!strcmp(argv[i], "-w")) {
             if ((i + 1) >= argc) {
-		fprintf(stderr, "missing argument for -w\n"); 
-		return -1; 
+		fprintf(stderr, "missing argument for -w\n");
+		return -1;
 	    }
 	    fiveminutes = atoi(argv[++i]);
 	} else if (!strcmp(argv[i], "-hr")) {
 	    int hr;
             if ((i + 1) >= argc) {
-		fprintf(stderr, "missing argument for -hr\n"); 
-		return -1; 
+		fprintf(stderr, "missing argument for -hr\n");
+		return -1;
 	    }
 	    hr = atoi(argv[++i]);
 	    if ((hr < 1) || (hr > 36)) {
@@ -1297,8 +1298,8 @@ ParseArgs(int argc, char *argv[])
 	else if (!strcmp(argv[i], "-cb")) {
 	    Sawcbs = 1;
             if ((i + 1) >= argc) {
-		fprintf(stderr, "missing argument for -cb\n"); 
-		return -1; 
+		fprintf(stderr, "missing argument for -cb\n");
+		return -1;
 	    }
 	    numberofcbs = atoi(argv[++i]);
 	    if ((numberofcbs < 10000) || (numberofcbs > 2147483647)) {
@@ -1310,8 +1311,8 @@ ParseArgs(int argc, char *argv[])
 	} else if (!strcmp(argv[i], "-busyat")) {
 	    Sawbusy = 1;
 	    if ((i + 1) >= argc) {
-		fprintf(stderr, "missing argument for -busyat\n"); 
-		return -1; 
+		fprintf(stderr, "missing argument for -busyat\n");
+		return -1;
 	    }
 	    busy_threshold = atoi(argv[++i]);
 	    if (busy_threshold < 10) {
@@ -1326,8 +1327,8 @@ ParseArgs(int argc, char *argv[])
 	else if (!strcmp(argv[i], "-m")) {
 	    extern int aixlow_water;
 	    if ((i + 1) >= argc) {
-		fprintf(stderr, "missing argument for -m\n"); 
-		return -1; 
+		fprintf(stderr, "missing argument for -m\n");
+		return -1;
 	    }
 	    aixlow_water = atoi(argv[++i]);
 	    if ((aixlow_water < 0) || (aixlow_water > 30)) {
@@ -1347,14 +1348,14 @@ ParseArgs(int argc, char *argv[])
 	    rxkadDisableDotCheck = 1;
 	} else if (!strcmp(argv[i], "-rxmaxmtu")) {
 	    if ((i + 1) >= argc) {
-		fprintf(stderr, "missing argument for -rxmaxmtu\n"); 
-		return -1; 
+		fprintf(stderr, "missing argument for -rxmaxmtu\n");
+		return -1;
 	    }
 	    rxMaxMTU = atoi(argv[++i]);
-	    if ((rxMaxMTU < RX_MIN_PACKET_SIZE) || 
+	    if ((rxMaxMTU < RX_MIN_PACKET_SIZE) ||
 		(rxMaxMTU > RX_MAX_PACKET_DATA_SIZE)) {
 		printf("rxMaxMTU %d%% invalid; must be between %d-%" AFS_SIZET_FMT "\n",
-		       rxMaxMTU, RX_MIN_PACKET_SIZE, 
+		       rxMaxMTU, RX_MIN_PACKET_SIZE,
 		       RX_MAX_PACKET_DATA_SIZE);
 		return -1;
 		}
@@ -1362,8 +1363,8 @@ ParseArgs(int argc, char *argv[])
 	    extern char local_realms[AFS_NUM_LREALMS][AFS_REALM_SZ];
 	    extern int  num_lrealms;
 	    if ((i + 1) >= argc) {
-		fprintf(stderr, "missing argument for -realm\n"); 
-		return -1; 
+		fprintf(stderr, "missing argument for -realm\n");
+		return -1;
 	    }
 	    if (strlen(argv[++i]) >= AFS_REALM_SZ) {
 		printf
@@ -1371,7 +1372,7 @@ ParseArgs(int argc, char *argv[])
 		     AFS_REALM_SZ);
 		return -1;
 	    }
-	    if (num_lrealms == -1) 
+	    if (num_lrealms == -1)
 		num_lrealms = 0;
 	    if (num_lrealms >= AFS_NUM_LREALMS) {
 		printf
@@ -1432,10 +1433,20 @@ ParseArgs(int argc, char *argv[])
 	else if (strcmp(argv[i], "-mrafslogs") == 0) {
 	    /* set syslog logging flag */
 	    mrafsStyleLogs = 1;
-	} 
+	}
 	else if (strcmp(argv[i], "-saneacls") == 0) {
 	    saneacls = 1;
 	}
+	else if (strcmp(argv[i], "-sync") == 0) {
+	    if ((i + 1) >= argc) {
+		printf("You have to specify -sync <sync behavior>\n");
+		return -1;
+	    }
+	    if (ih_SetSyncBehavior(argv[++i])) {
+		printf("Invalid -sync value %s\n", argv[i]);
+		return -1;
+	    }
+	}
 	else {
 	    return (-1);
 	}
@@ -1481,6 +1492,12 @@ ParseArgs(int argc, char *argv[])
     if (auditFileName)
 	osi_audit_file(auditFileName);
 
+    lwps_max = max_fileserver_thread() - FILESERVER_HELPER_THREADS;
+    if (lwps > lwps_max)
+	lwps = lwps_max;
+    else if (lwps < 6)
+	lwps = 6;
+
     return (0);
 
 }				/*ParseArgs */
@@ -1492,12 +1509,13 @@ static void
 NewParms(int initializing)
 {
     static struct afs_stat sbuf;
-    register int i, fd;
+    int i, fd;
     char *parms;
     char *argv[MAXPARMS];
-    register int argc;
+    int argc;
 
     if (!(afs_stat("/vice/file/parms", &sbuf))) {
+	ViceLog(0, ("/vice/file/parms is deprecated, and will be removed in a future release!"));
 	parms = (char *)malloc(sbuf.st_size);
 	if (!parms)
 	    return;
@@ -1560,7 +1578,7 @@ void
 Die(char *msg)
 {
     ViceLog(0, ("%s\n", msg));
-    assert(0);
+    osi_Panic("%s\n", msg);
 
 }				/*Die */
 
@@ -1582,7 +1600,7 @@ InitPR(void)
     }
 
 #ifdef AFS_PTHREAD_ENV
-    assert(pthread_key_create(&viced_uclient_key, NULL) == 0);
+    osi_Assert(pthread_key_create(&viced_uclient_key, NULL) == 0);
 #endif
 
     SystemId = SYSADMINID;
@@ -1641,6 +1659,11 @@ vl_Initialize(const char *confDir)
 	exit(1);
     }
     code = afsconf_GetCellInfo(tdir, NULL, AFSCONF_VLDBSERVICE, &info);
+    if (code) {
+	ViceLog(0,
+		("vl_Initialize: Failed to get cell information\n"));
+	exit(1);
+    }
     if (info.numServers > MAXSERVERS) {
 	ViceLog(0,
 		("vl_Initialize: info.numServers=%d (> MAXSERVERS=%d)\n",
@@ -1809,16 +1832,16 @@ WriteSysIdFile(void)
 }
 
 /*
- * defect 10966 
+ * defect 10966
  * This routine sets up the buffers for the VL_RegisterAddrs RPC. All addresses
  * in FS_HostAddrs[] are in NBO, while the RPC treats them as a "blob" of data
- * and so we need to convert each of them into HBO which is what the extra 
+ * and so we need to convert each of them into HBO which is what the extra
  * array called FS_HostAddrs_HBO is used here.
  */
 static afs_int32
 Do_VLRegisterRPC(void)
 {
-    register int code;
+    int code;
     bulkaddrs addrs;
     afs_uint32 FS_HostAddrs_HBO[ADDRSPERSITE];
     int i = 0;
@@ -1860,8 +1883,8 @@ SetupVL(void)
 
     if (AFSDIR_SERVER_NETRESTRICT_FILEPATH || AFSDIR_SERVER_NETINFO_FILEPATH) {
 	/*
-	 * Find addresses we are supposed to register as per the netrestrict 
-	 * and netinfo files (/usr/afs/local/NetInfo and 
+	 * Find addresses we are supposed to register as per the netrestrict
+	 * and netinfo files (/usr/afs/local/NetInfo and
 	 * /usr/afs/local/NetRestict)
 	 */
 	char reason[1024];
@@ -1881,7 +1904,7 @@ SetupVL(void)
 
     if (FS_HostAddr_cnt == 1 && rxBind == 1)
 	code = FS_HostAddrs[0];
-    else 
+    else
 	code = htonl(INADDR_ANY);
     return code;
 }
@@ -1898,7 +1921,7 @@ InitVL(void)
     code = vl_Initialize(AFSDIR_SERVER_ETC_DIRPATH);
     if (code != 0) {
 	ViceLog(0,
-		("Couldn't initialize protection library; code=%d.\n", code));
+		("Couldn't initialize volume location library; code=%d.\n", code));
 	return code;
     }
 
@@ -1966,9 +1989,6 @@ main(int argc, char *argv[])
 		argv[0]);
 	exit(2);
     }
-#ifndef AFS_QUIETFS_ENV
-    console = afs_fopen("/dev/console", "w");
-#endif
     /* set ihandle package defaults prior to parsing args */
     ih_PkgDefaults();
 
@@ -1977,7 +1997,7 @@ main(int argc, char *argv[])
 	exit(-1);
     }
 #ifdef AFS_PTHREAD_ENV
-    assert(pthread_mutex_init(&fileproc_glock_mutex, NULL) == 0);
+    MUTEX_INIT(&fileproc_glock_mutex, "fileproc", MUTEX_DEFAULT, 0);
 #endif /* AFS_PTHREAD_ENV */
 
 #ifdef AFS_SGI_VNODE_GLUE
@@ -2007,11 +2027,7 @@ main(int argc, char *argv[])
 	ViceLog(0, ("Both -spare and -pctspare specified, exiting.\n"));
 	exit(-1);
     }
-#ifdef AFS_SGI_XFS_IOPS_ENV
-    ViceLog(0, ("XFS/EFS File server starting\n"));
-#else
-    ViceLog(0, ("File server starting\n"));
-#endif
+    LogCommandLine(argc, argv, "starting", "", "File server", FSLog);
 
 #if defined(AFS_PTHREAD_ENV) && !defined(AFS_NT40_ENV)
     /* initialize the pthread soft signal handler thread */
@@ -2032,7 +2048,7 @@ main(int argc, char *argv[])
     nice(-5);			/* TODO: */
 #endif
 #endif
-    assert(DInit(buffs) == 0);
+    osi_Assert(DInit(buffs) == 0);
 #ifdef AFS_DEMAND_ATTACH_FS
     FS_STATE_INIT;
 #endif
@@ -2046,8 +2062,8 @@ main(int argc, char *argv[])
 #endif
     CheckAdminName();
 
-    /* if we support more than 16 threads, then we better have the ability 
-     ** to keep open a large number of files simultaneously 
+    /* if we support more than 16 threads, then we better have the ability
+     ** to keep open a large number of files simultaneously
      */
 #if	defined(AFS_AIX_ENV) && !defined(AFS_AIX42_ENV)
     curLimit = OPEN_MAX;	/* for pre AIX 4.2 systems */
@@ -2085,7 +2101,7 @@ main(int argc, char *argv[])
 		 curLimit, lwps, vol_io_params.fd_max_cachesize));
     }
 #ifndef AFS_PTHREAD_ENV
-    assert(LWP_InitializeProcessSupport(LWP_MAX_PRIORITY - 2, &parentPid) ==
+    osi_Assert(LWP_InitializeProcessSupport(LWP_MAX_PRIORITY - 2, &parentPid) ==
 	   LWP_SUCCESS);
 #endif /* !AFS_PTHREAD_ENV */
 
@@ -2142,7 +2158,7 @@ main(int argc, char *argv[])
     afsconf_BuildServerSecurityObjects(confDir, AFSCONF_SEC_OBJS_RXKAD_CRYPT,
 				       &securityClasses, &numClasses);
 
-    tservice = rx_NewServiceHost(rx_bindhost,  /* port */ 0, /* service id */ 
+    tservice = rx_NewServiceHost(rx_bindhost,  /* port */ 0, /* service id */
 				 1,	/*service name */
 				 "AFS",
 				 securityClasses, numClasses,
@@ -2205,15 +2221,13 @@ main(int argc, char *argv[])
     }
 
     /* allow super users to manage RX statistics */
-    rx_SetRxStatUserOk(fs_rxstat_userok);
+    rx_SetRxStatUserOk(viced_SuperUser);
 
-#ifdef AFS_PTHREAD_ENV
-    assert(pthread_cond_init(&fsync_cond, NULL) == 0);
-    assert(pthread_mutex_init(&fsync_glock_mutex, NULL) == 0);
-#endif
+    CV_INIT(&fsync_cond, "fsync", CV_DEFAULT, 0);
+    MUTEX_INIT(&fsync_glock_mutex, "fsync", MUTEX_DEFAULT, 0);
 
 #if !defined(AFS_DEMAND_ATTACH_FS)
-    /* 
+    /*
      * For DAFS, we do not start the Rx server threads until after
      * the volume package is initialized, and fileserver state is
      * restored.  This is necessary in order to keep host and callback
@@ -2226,11 +2240,11 @@ main(int argc, char *argv[])
 #endif
 
     /* we ensure that there is enough space in the vnode buffer to satisfy
-     ** requests from all concurrent threads. 
+     ** requests from all concurrent threads.
      ** the maximum number of vnodes used by a single thread at any one time
      ** is three ( "link" uses three vnodes simultaneously, one vLarge and
      ** two vSmall for linking files and two vLarge and one vSmall for linking
-     ** files  ) : dhruba 
+     ** files  ) : dhruba
      */
     minVnodesRequired = 2 * lwps + 1;
     if (minVnodesRequired > nSmallVns) {
@@ -2246,13 +2260,14 @@ main(int argc, char *argv[])
 
     /* We now do this after getting the listener up and running, so that client
      * connections don't timeout (maybe) if a file server is restarted, since it
-     * will be available "real soon now".  Worry about whether we can satisfy the 
+     * will be available "real soon now".  Worry about whether we can satisfy the
      * calls in the volume package itself.
      */
     VOptDefaults(fileServer, &opts);
     opts.nLargeVnodes = large;
     opts.nSmallVnodes = nSmallVns;
     opts.volcache = volcache;
+    opts.unsafe_attach = unsafe_attach;
 
     if (VInitVolumePackage2(fileServer, &opts)) {
 	ViceLog(0,
@@ -2295,41 +2310,33 @@ main(int argc, char *argv[])
 
 #ifdef AFS_PTHREAD_ENV
     ViceLog(5, ("Starting pthreads\n"));
-    assert(pthread_attr_init(&tattr) == 0);
-    assert(pthread_attr_setdetachstate(&tattr, PTHREAD_CREATE_DETACHED) == 0);
+    osi_Assert(pthread_attr_init(&tattr) == 0);
+    osi_Assert(pthread_attr_setdetachstate(&tattr, PTHREAD_CREATE_DETACHED) == 0);
 
-    assert(pthread_create
+    osi_Assert(pthread_create
 	   (&serverPid, &tattr, FiveMinuteCheckLWP,
 	    &fiveminutes) == 0);
-    assert(pthread_create
+    osi_Assert(pthread_create
 	   (&serverPid, &tattr, HostCheckLWP, &fiveminutes) == 0);
-    assert(pthread_create
+    osi_Assert(pthread_create
 	   (&serverPid, &tattr, FsyncCheckLWP, &fiveminutes) == 0);
 #else /* AFS_PTHREAD_ENV */
     ViceLog(5, ("Starting LWP\n"));
-    assert(LWP_CreateProcess
+    osi_Assert(LWP_CreateProcess
 	   (FiveMinuteCheckLWP, stack * 1024, LWP_MAX_PRIORITY - 2,
 	    (void *)&fiveminutes, "FiveMinuteChecks",
 	    &serverPid) == LWP_SUCCESS);
 
-    assert(LWP_CreateProcess
+    osi_Assert(LWP_CreateProcess
 	   (HostCheckLWP, stack * 1024, LWP_MAX_PRIORITY - 2,
 	    (void *)&fiveminutes, "HostCheck", &serverPid) == LWP_SUCCESS);
-    assert(LWP_CreateProcess
+    osi_Assert(LWP_CreateProcess
 	   (FsyncCheckLWP, stack * 1024, LWP_MAX_PRIORITY - 2,
 	    (void *)&fiveminutes, "FsyncCheck", &serverPid) == LWP_SUCCESS);
 #endif /* AFS_PTHREAD_ENV */
 
     FT_GetTimeOfDay(&tp, 0);
 
-#ifndef AFS_QUIETFS_ENV
-    if (console != NULL) { 
-        time_t t = tp.tv_sec;
-	fprintf(console, "File server has started at %s\r",
-		afs_ctime(&t, tbuffer, sizeof(tbuffer)));
-    }
-#endif
-
     /*
      * Figure out the FileServer's name and primary address.
      */
@@ -2366,7 +2373,7 @@ main(int argc, char *argv[])
 	sleep(1000);		/* long time */
     }
 #else /* AFS_PTHREAD_ENV */
-    assert(LWP_WaitProcess(&parentPid) == LWP_SUCCESS);
+    osi_Assert(LWP_WaitProcess(&parentPid) == LWP_SUCCESS);
 #endif /* AFS_PTHREAD_ENV */
     return 0;
 }
